#!/usr/bin/perl

#
# "Assembler source converter" (C) 28-Jul-2020y by *HW*
#
#   This is the converter of assembler sources
#   from a some unknown internal representation
#   to a free text form which can be further compiled
#   using other assembler (cross- or more modern)
#

# Instruction tables are from
# https://github.com/lllllzzzzz/Intel-8080-Disassembler/blob/master/disassembler.c
# They contained a lot of bugs, those I found I fixed
#

@INSTR_SZ = (
#   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
    1, 3, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1,
    1, 3, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1,
    1, 3, 3, 1, 1, 1, 2, 1, 1, 1, 3, 1, 1, 1, 2, 1,
    1, 3, 3, 1, 1, 1, 2, 1, 1, 1, 3, 1, 1, 1, 2, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 3, 3, 3, 1, 2, 1, 1, 1, 3, 1, 3, 3, 2, 1,
    1, 1, 3, 2, 3, 1, 2, 1, 1, 1, 3, 2, 3, 1, 2, 1,
    1, 1, 3, 1, 3, 1, 2, 1, 1, 1, 3, 1, 3, 3, 2, 1,
    1, 1, 3, 1, 3, 1, 2, 1, 1, 1, 3, 1, 3, 3, 2, 1
);
@INSTR = (
#     0          1          2          3          4          5           6          7          8          9          A          b          c          d          e          f
    "nop",     "lxi b,",  "stax b",  "inx b",   "inr b",   "dcr b",    "mvi b,",  "rlc",     "illegal", "dad b",   "ldax b",  "dcx b",   "inr c",   "dcr c",   "mvi c,",  "rrc",
    "illegal", "lxi d,",  "stax d",  "inx d",   "inr d",   "dcr d",    "mvi d,",  "ral",     "illegal", "dad d",   "ldax d",  "dcx d",   "inr e",   "dcr e",   "mvi e,",  "rar",
    "illegal", "lxi h,",  "shld",    "inx h",   "inr h",   "dcr h",    "mvi h,",  "daa",     "illegal", "dad h",   "lhld",    "dcx h",   "inr l",   "dcr l",   "mvi l,",  "cma",
    "illegal", "lxi sp,", "sta",     "inx sp",  "inr m",   "dcr m",    "mvi m,",  "stc",     "illegal", "dad sp",  "lda",     "dcx sp",  "inr a",   "dcr a",   "mvi a,",  "cmc",
    "mov b,b", "mov b,c", "mov b,d", "mov b,e", "mov b,h", "mov b,l",  "mov b,m", "mov b,a", "mov c,b", "mov c,c", "mov c,d", "mov c,e", "mov c,h", "mov c,l", "mov c,m", "mov c,a",
    "mov d,b", "mov d,c", "mov d,d", "mov d,e", "mov d,h", "mov d,l",  "mov d,m", "mov d,a", "mov e,b", "mov e,c", "mov e,d", "mov e,e", "mov e,h", "mov e,l", "mov e,m", "mov e,a",
    "mov h,b", "mov h,c", "mov h,d", "mov h,e", "mov h,h", "mov h,l",  "mov h,m", "mov h,a", "mov l,b", "mov l,c", "mov l,d", "mov l,e", "mov l,h", "mov l,l", "mov l,m", "mov l,a",
    "mov m,b", "mov m,c", "mov m,d", "mov m,e", "mov m,h", "mov m,l",  "hlt",     "mov m,a", "mov a,b", "mov a,c", "mov a,d", "mov a,e", "mov a,h", "mov a,l", "mov a,m", "mov a,a",
    "add b",   "add c",   "add d",   "add e",   "add h",   "add l",    "add m",   "add a",   "adc b",   "adc c",   "adc d",   "adc e",   "adc h",   "adc l",   "adc m",   "adc a",
    "sub b",   "sub c",   "sub d",   "sub e",   "sub h",   "sub l",    "sub m",   "sbb a",   "sbb b",   "sbb c",   "sbb d",   "sbb e",   "sbb h",   "sbb l",   "sbb m",   "sbb a",
    "ana b",   "ana c",   "ana d",   "ana e",   "ana h",   "ana l",    "ana m",   "ana a",   "xra b",   "xra c",   "xra d",   "xra e",   "xra h",   "xra l",   "xra m",   "xra a",
    "ora b",   "ora c",   "ora d",   "ora e",   "ora h",   "ora l",    "ora m",   "ora a",   "cmp b",   "cmp c",   "cmp d",   "cmp e",   "cmp h",   "cmp l",   "cmp m",   "cmp a",
    "rnz",     "pop b",   "jnz",     "jmp",     "cnz",     "push b",   "adi",     "rst 0",   "rz",      "ret",     "jz",      "illegal", "cz",      "call",    "aci",     "rst 1",
    "rnc",     "pop d",   "jnc",     "out",     "cnc",     "push d",   "sui",     "rst 2",   "rc",      "illegal", "jc",      "in",      "cc",      "illegal", "sbi",     "rst 3",
    "rpo",     "pop h",   "jpo",     "xthl",    "cpo",     "push h",   "ani",     "rst 4",   "rpe",     "pchl",    "jpe",     "xchg",    "cpe",     "illegal", "xri",     "rst 5",
    "rp",      "pop psw", "jp",      "di",      "cp",      "push psw", "ori",     "rst 6",   "rm",      "sphl",    "jm",      "ei",      "cm",      "illegal", "cpi",     "rst 7"
);

sub instr {
    my ($code) = ord($_[0]);
    my ($sz) = $INSTR_SZ[$code];
    my ($insn) = $INSTR[$code];

    return "DB" if $code == 0x08;
    return "DW " if $code == 0x30;
    return ".ORG " if $code == 0x10;
    return ".EOF" if $code == 0xFD;

    unless ($insn =~ s{^(\S+) }{substr($1 . (' ' x 10), 0, 10)}e) {
        $insn = substr($insn . (' ' x 10), 0, 10);
    }
    $insn .= ' ' if $insn =~ m{,$};

    return uc($insn);
}

# This is O32.LV0 from warez-d4.rar\ARCHIVE\SYSTEM.BIN\O32.LV0 
$fname = ($ARGV[0] or "O32.LV0");
open F,"<$fname" or die "$fname - $!";
binmode F; read F, $d, -s F; close F;
substr($d, 0, 0x16) = "; CONVERTED\r";

$d =~ s{\x0D([A-Za-z0-9\x20]+:)(.)}{ "\n" . $1 . " " . instr($2) }esg;
$d =~ s{\x0D([\x09])(.)}{"\n" . (" " x 8) . instr($2)}esg;

$d =~ s{\s+$}{}mg;
$d =~ s{(.EOF).*}{\1}s;

# $d =~ s{([\x60-\x7F])}{ # OEM CP 866
#     substr("\x9E\x80\x81\x96\x84\x85\x94\x83"
#            ."\x95\x88\x89\x8A\x8B\x8C\x8D\x8E"
#            ."\x8F\x9F\x90\x91\x92\x93\x86\x82"
#            ."\x9C\x9B\x87\x98\x9D\x99\x97\x9A"
#            , ord($1) - 0x60, 1)
# }eg;

$d =~ s{([\x60-\x7F])}{ # ANSI CP Windows 1251
    substr("\xDE\xC0\xC1\xD6\xC4\xC5\xD4\xC3"
           ."\xD5\xC8\xC9\xCA\xCB\xCC\xCD\xCE"
           ."\xCF\xDF\xD0\xD1\xD2\xD3\xC6\xC2"
           ."\xDC\xDB\xC7\xD8\xDD\xD9\xD7\xDA"
           , ord($1) - 0x60, 1)
}eg;

print while <DATA>;
print $d;
__END__
; PREAMBLE

@KY* EQU 0F806H
@TTY EQU 0F809H
@INK EQU 0F812H
@CUR EQU 0F82DH

$TC EQU 0BE36H
$VC EQU 0BE3DH
$CX EQU 0BE3EH

